/**
 * jparser.h -- Parser header file.
 *
 * Author: Gaurav Vaish <gvaish@iitk.ac.in>
 *         Gopal V (gopalv82@yahoo.com)
 *
 * (C) Gaurav Vaish, Gopal V; 2002
 */

#ifndef __JPARSER_H
#define __JPARSER_H

#include <jopcodes.h>
#include <defines.h>
#include <stdio.h>

typedef struct
{
	byte*  data;
	UInt32 next;
	UInt32 left;
} jbuffer;

/**
 * Constant Pool Entries : Values
 */

#define UTF8                1
#define INTEGER             3
#define FLOAT               4
#define LONG                5
#define DOUBLE              6
#define CLASSREF            7
#define STRING              8
#define FIELDREF            9
#define METHODREF          10
#define INTERFACEMETHODREF 11
#define NAMEANDTYPE        12

/**
 * Attribute Types
 */

#define ATTRIBUTE_TOP    0
#define ATTRIBUTE_FIELD  1
#define ATTRIBUTE_METHOD 2
#define ATTRIBUTE_CODE   3

/**
 * Access Type (Flags)
 */

#define ACCESS_PUBLIC       0x0001
#define ACCESS_PRIVATE      0x0002
#define ACCESS_PROTECTED    0x0004
#define ACCESS_STATIC       0x0008
#define ACCESS_FINAL        0x0010
#define ACCESS_SYNCHRONIZED 0x0020
#define ACCESS_VOLATILE     0x0040
#define ACCESS_TRANSIENT    0x0080
#define ACCESS_NATIVE       0x0100
#define ACCESS_ABSTRACT     0x0400

/**
 * Java Constant Pool Entry structure
 */

typedef struct
{
	UInt8  type;
	UInt16 length;
	UInt8  skip;
	union
	{
		char   *utf8string;
		UInt32  intvalue;
		UInt64  longvalue;
		float   floatvalue;
		double  doublevalue;
		UInt32  strvalue;
		struct
		{
			UInt32  name_index;
			char   *full_name;
		} class_value;
		struct
		{
			UInt32  class_index;
			UInt32  name_type;
			char   *full_class;
			char   *full_name;
			char   *full_type;
		} ref_value;
		struct
		{
			UInt32  name_index;
			UInt32  type_index;
			char   *full_name;
			char   *full_type;
		} name_type;
	} values;
} JVCPEntry;

/**
 * Java Exception Handler structure
 */

typedef struct
{
	UInt16 start_pc;
	UInt16 end_pc;
	UInt16 handler_pc;
	UInt16 catch_type;
} JVExceptionHandler;

/**
 * Java Exception structure
 */

typedef struct
{
	UInt16  class_index;
	char   *class_name;
} JVException;

/**
 * Java Attribute Entry structure
 */

typedef struct
{
	UInt16   name_index;
	UInt32   byte_count;
	UInt8    type;
	jbuffer *buffer;
	union
	{
		struct
		{
			UInt16  source_index;
			char   *full_source;
		} source;
		struct
		{
			UInt16       except_count;
			JVException *excepts;
		} exception;
		struct
		{
			UInt16              max_stack;
			UInt16              max_locals;
			UInt32              code_count;
			byte               *code;
			UInt16              handler_count;
			JVExceptionHandler *handlers;
			UInt16              attr_count;
			byte               *attributes;
		} code;
	} attributes;
	char *full_name;
} JVAttributeEntry;

/**
 * Java Field Entry structure
 */

typedef struct
{
	UInt16            acc_flags;
	UInt16            name_index;
	UInt16            type_index;
	UInt16            attr_count;
	JVAttributeEntry *attr_entries;
	char             *full_flags;
	char             *full_name;
	char             *full_type;
} JVFieldEntry;

/**
 * Java Interface Entry structure
 */

typedef struct
{
	UInt16  index;
	char   *full_name;
} JVInterfaceEntry;

/**
 * Java Method Entry structure
 */

typedef struct
{
	UInt16            acc_flags;
	UInt16            name_index;
	UInt16            type_index;
	UInt16            attr_count;
	JVAttributeEntry *attr_entries;
	UInt16            code_index;
	UInt16            except_index;
	char             *full_flags;
	char             *full_name;
	char             *full_type;
} JVMethodEntry;

/**
 * Java Class structure
 */

typedef struct
{
	UInt32            magic;
	UInt16            version_minor;
	UInt16            version_major;
	UInt16            cp_count;
	JVCPEntry        *cp_entries;
	UInt16            acc_flags;
	UInt16            this_class;
	UInt16            super_class;
	UInt16            interface_count;
	JVInterfaceEntry *interface_entries;
	UInt16            field_count;
	JVFieldEntry     *field_entries;
	UInt16            method_count;
	JVMethodEntry    *method_entries;
	UInt16            attr_count;
	JVAttributeEntry *attr_entries;
	char             *full_name;
	char             *full_super;
	char             *full_flags;
} JVClass;

/**
 * File reader methods
 */

inline UInt8  read_byte(FILE *fp);
inline UInt16 read_uint16(FILE *fp);
inline UInt32 read_uint32(FILE *fp);
inline UInt64 read_uint64(FILE *fp);

inline void   skip_bytes(FILE *fp, int count);

inline byte buffer_read_byte(jbuffer *buffer);
inline UInt16 buffer_read_uint16(jbuffer *buffer);
inline UInt32 buffer_read_uint32(jbuffer *buffer);

void buffer_skip_bytes(jbuffer *buffer, UInt32 count);

/**
 * Parser Methods (provide values to JVClass structure entries)
 */

int read_cp_entry(FILE *fp, JVCPEntry *cp_entry);
int read_cp(FILE *fp, JVClass *klass);
int read_flags(FILE *fp, JVClass *klass);
int read_interfaces(FILE *fp, JVClass *klass);

JVAttributeEntry* read_attributes(FILE *fp, int count, int type);
int read_fields(FILE *fp, JVClass *klass);
int read_methods(FILE *fp, JVClass *klass);
int read_class(FILE *fp, JVClass *klass);

/**
 * Resolving methods (resolve all entries in the class)
 */

char* resolve_utf8(UInt16 index, JVCPEntry *cp_entry);
char* resolve_string(UInt16 index, JVCPEntry *cp_entry);
char* resolve_classref(UInt16 index, JVCPEntry *cp_entry);

JVCPEntry* resolve_name_type(UInt16 index, JVCPEntry *cp_entry);
JVCPEntry* resolve_ref(UInt16 index, JVCPEntry *cp_entry);

int resolve_cp(JVClass *klass, JVCPEntry *cp_entry);
char* resolve_flags(UInt16 flags);

int resolve_all_interfaces(JVClass *klass, JVCPEntry *cp_entry);
int resolve_top_attributes(JVClass *klass, JVCPEntry *cp_entry);
int resolve_field(JVFieldEntry *field_entry, JVCPEntry *cp_entry);
int resolve_all_fields(JVClass *klass,JVCPEntry *cp_entry);
int resolve_all_method_attributes(JVMethodEntry *method,JVCPEntry *cp_entry);
int resolve_method(JVMethodEntry *method,JVCPEntry *cp_entry);
int resolve_all_methods(JVClass *klass,JVCPEntry *cp_entry);
int resolve_class(JVClass *klass,JVCPEntry *cp_entry);

int process_source_attribute(JVAttributeEntry *attr_entry, JVCPEntry *cp_entry);
int process_methodcode_attribute(JVAttributeEntry *code, JVCPEntry *cp_entry);

#endif